# Script ler_faturas_CPFL v1.0 - 16/11/2021
# Elaboracao: Daniel Hoffman
# Testes: Fernando Finger

# Clean residual data
rm(list = ls())
# library need for using pipes
library(dplyr)
# match list for month abreviations
meses <- c("jan", "fev", "mar", "abr", "mai", "jun", "jul", "ago", "set" , "out", "nov" , "dez" )
#########################################################################################
# get .pdf directory - will refresh until selection of any directory
tcltk::tkmessageBox(title = "AVISO!", message = paste("Selecione a pasta raiz que contem os arquivos .PDF: "), icon = "info", type = "ok")
repeat{
    Pasta_Arquivos_PDF <- tcltk::tk_choose.dir("Escolher pasta para exploracao")
    if(!is.na(Pasta_Arquivos_PDF)){
        break
    }
}
# build filter for Excel files
filters <- matrix(c("XLSX files", ".xlsx"), 2, 2, byrow = TRUE)
# get xlsx file path though canceladion/non selection will not cause refresh!
tcltk::tkmessageBox(title = "AVISO!", message = paste("Selecione o arquivo de dados ja existente. Caso nao exista, clique em cancelar."), icon = "info", type = "ok")
repeat{
    Excel_Output <- tcltk::tk_choose.files(filter = filters)
	# if user cancels/closes screen set variable for later use
    if(length(Excel_Output) == 0){
	    Excel_Output <- "nada"
	    break
    }  
	# if user selected a file check if file is in use and abort script in this case
    if(!is.na(Excel_Output)){
    	# check by searching for files with same name in directory (bkp file should have only diferent preceding characters 
        if(sum(stringr::str_detect(list.files(dirname(Excel_Output)), 
                                   stringr::str_remove(Excel_Output, paste0(dirname(Excel_Output),"/")))) != 1){
			# send out message
            tcltk::tkmessageBox(title = "PROBLEMA ;(", message = "Ou Planilha esta aberta ou nao existe?!", icon = "info", type = "ok")
            # causes script to fail
            UNDECLARED()
        }
        break
    }
}
tryCatch(
	expr = {
        # check if file exists
        if(file.exists(Excel_Output)){
            # check if excel file has at least on worksheet with fatura data
            if(sum(readxl::excel_sheets(path = Excel_Output) %in% c("fat_A", "fat_B")) > 0){
		        # check if sheet for fatura type A exists and read it else declare variable as NULL
		        if(sum(readxl::excel_sheets(path = Excel_Output) == "fat_A") == 1){
			        read_data_A <- openxlsx::read.xlsx(Excel_Output, sheet = "fat_A") %>% 
                        dplyr::mutate(across(c(17:19,21,22), ~as.Date(.x, origin = "1899-12-30")))
			        read_files_A <- unique(read_data_A$.id)
		        }else{
			        # declare empty variable for later use
			        read_data_A <- NULL
			        read_files_A <- NULL
		        }
		        # check if sheet for fatura type B exists and read it else declare variable as NULL
		        if(sum(readxl::excel_sheets(path = Excel_Output) == "fat_B") == 1){
			        read_data_B <- openxlsx::read.xlsx(Excel_Output, sheet = "fat_B") %>% 
                        dplyr::mutate(across(16:20, ~as.Date(.x, origin = "1899-12-30")))
			        read_files_B <- unique(read_data_B$.id)
		        } else {
			        # declare empty variable for later use
			        read_data_B <- NULL
			        read_files_B <- NULL
		        }
		        # combine all paths to one vector for later use
	            read_files <- c(read_files_A, read_files_B)
            }else{
                # msg where new output file will be created and create path plus empty variables
                tcltk::tkmessageBox(title = "AVISO!", message = paste("Sera criado novo arquivo em:\n", Pasta_Arquivos_PDF), icon = "info", type = "ok")
	            Excel_Output <- paste0(Pasta_Arquivos_PDF, "/output.xlsx")
                read_data_A <- NULL
	            read_data_B <- NULL
                read_files <- NULL
            }
		}else{
	        Excel_Output <- paste0(Pasta_Arquivos_PDF, "/output.xlsx")
            # msg where new output file will be created and create path plus empty variables
            tcltk::tkmessageBox(title = "AVISO!", message = paste("Sera criado novo arquivo em:\n", Excel_Output), icon = "info", type = "ok")
            read_data_A <- NULL
	        read_data_B <- NULL
            read_files <- NULL
        }
	},error = function(e){ 
    msgBox <- tcltk::tkmessageBox(title = "PROBLEMA ;(",
                                  message = "Algo no arquivo Excel parece estar errado - tente remover-lo e execute a leitura novamente.", icon = "info", type = "ok")
    UNDECLARED()
  }
)
# one could check if the excel files have the needed columns names and data as this could generate and error
#########################################################################################
# list all .pdf files wih full names and directories
files <- list.files(Pasta_Arquivos_PDF, full.names = TRUE, recursive = TRUE, include.dirs = TRUE)
# reduce to only .pdf files with path
pdf_files <- data.frame(path = files[stringr::str_detect(files, pattern = "\\.pdf$")]) 
# files that have not previously been read
pdf_files_to_read <- pdf_files %>% 
    dplyr::filter(!path %in% read_files) %>% 
    dplyr::pull(path) %>%
	stringr::str_replace("//", "/")
# avisar sobre pre leitura
tcltk::tkmessageBox(title = "AVISO!", message = paste("Foram encontrados:", nrow(pdf_files), "arquivos de PDF!\nSendo", length(pdf_files_to_read), "novos!"), icon = "info", type = "ok")
# function to read in file, read lines and drop leading and trailing blanks
read_pdfs <- function(file){
    # 
    pdf_text <- pdftools::pdf_text(file)[[1]] %>% 
        readr::read_lines() %>%
        trimws()
    return(pdf_text) 
}
# read in all files... possibly need to be more error prone
read_pdfs <- purrr::possibly(read_pdfs,"0000000000")
pdf_texts <- purrr::map(pdf_files_to_read, ~read_pdfs(.x)) 
# pdf function to see if it is type A, B or Segunda Via (if all zero then none of the prior)
judge_pdfs <- function(pdf_text){  
    # check what typ of pdf it is
    val <- data.frame(A = sum(stringr::str_detect(stringr::str_to_lower(pdf_text), "classifica.+a4\\s+")),
                      B = sum(stringr::str_detect(stringr::str_to_lower(pdf_text), "classifica.+b\\d.*")),
                      S = sum(stringr::str_detect(stringr::str_to_lower(pdf_text), "segunda via")))
    return(val)
}
# make custom function failure proof (in case of all 0)
judge_pdfs <- purrr::possibly(judge_pdfs, otherwise = data.frame(B = 0, A = 0, S = 0))
# apply rename and bind to df
ret <- purrr::map(pdf_texts, ~judge_pdfs(.x))
names(ret) <- pdf_files_to_read
valuation <- plyr::ldply(ret) 
# all segunda vias
segunda_via <- data.frame(path = valuation[valuation$S != 0, 1]) %>%
    dplyr::mutate(motivo = "segunda via")
# not relevant pdfs
nao_e_fatura <- data.frame(path = valuation[valuation$A == 0 & valuation$B == 0, 1]) %>%
    dplyr::mutate(motivo = "nao eh fatura")
# all relevant pdfs
towork <- valuation[valuation$S == 0 & (valuation$A != 0 | valuation$B != 0), ]
# all group A faturas
grupo_a <- towork[towork$A != 0, 1]
# all group B faturas
grupo_b <- towork[towork$B != 0, 1]
# inform numerical resume
tcltk::tkmessageBox(title = "AVISO!", message = paste("Faturas grupo A:", length(grupo_a), "\nFaturas grupo B:", length(grupo_b),"\nSegunda Via:", nrow(segunda_via), "\nNao sao Faturas:", nrow(nao_e_fatura)), icon = "info", type = "ok")
# abort by calling undeclared variable in case no new work is found (preserve last run with usefull data)
if(nrow(towork) == 0){
	tcltk::tkmessageBox(title = "AVISO!", message = paste("Nenhuma fatura nova para processar!"), icon = "info", type = "ok")
    UNDECLARED
}
#
tcltk::tkmessageBox(title = "AVISO!", message = paste("Por favor, clique OK e aguarde. Voce sera notificado assim que a leitura terminar!"), icon = "info", type = "ok")
#########################################################################################
# helper function to detect, clean and separte regex data
get_regex_2 <- function(txt_lines, my_regex_1, my_regex_2){
    txt_lines %>%
		dplyr::filter(stringr::str_detect(lines, pattern = my_regex_1)) %>%
		dplyr::pull(lines) %>%
		trimws() %>%
		stringr::str_extract(pattern = my_regex_2)
}
# 
get_regex_3 <- function(dt, p){
    ret <- dt %>%
	    dplyr::filter(y < 200) %>%
        dplyr::mutate(parser = readr::parse_number(stringr::str_remove_all(stringr::str_extract(lines, pattern = p), pattern = "\\."))) %>%
        dplyr::filter(!is.na(parser))
    if(nrow(ret) == 0){
        ret <- NA_real_
    }else{
        ret <- ret[[1,3]]
    }
}
# funcion to get column, change name and round row number down to nearest 10
get_column_by_position_5 <- function(table_data, x1, x2, varname){
    table_data %>% 
        dplyr::filter(x > x1 & x < x2) %>%
		dplyr::mutate(y = plyr::round_any(y, accuracy = 5, f = floor),
					  text = trimws(stringr::str_remove_all(text, pattern = "~"))) %>%
		dplyr::group_by(y) %>%
		dplyr::select(y, !!varname := text) %>%
		dplyr::slice_head(n = 1)
}
# funcion to get column, change name and round row number down to nearest 5
get_column_by_position_10 <- function(table_data, x1, x2, varname){
    table_data %>% 
        dplyr::filter(x > x1 & x < x2) %>%
		dplyr::mutate(y = plyr::round_any(y, accuracy = 10, f = floor),
					  text = trimws(stringr::str_remove_all(text, pattern = "~"))) %>%
		dplyr::group_by(y) %>%
		dplyr::select(y, !!varname := text) %>%
		dplyr::slice_head(n = 1)
}
# function to detect and label relevant info
get_descprod_with_info_CPFL_A <- function(data, sear, info){
    data %>% 
        dplyr::filter(stringr::str_detect(stringr::str_to_lower(DescProd), pattern = sear)) %>% 
	    dplyr::mutate(Msg = info) %>%
	    dplyr::select(.id, Referencia, UC, Msg, VlrTot)			
}
#########################################################################################
tryCatch(
	expr = {
	    # only process type A if any is there to read
	    if(length(grupo_a) != 0){
    	    # main function to extract data
    	    explore_fatura_a <- function(arquivo){
        	    # read data from pdf file
			    pdfs <- pdftools::pdf_data(arquivo)
			    # filter table region from first page of pdf doc
        	    table_data <- pdfs[[1]] %>% 
            	    dplyr::filter(y > 300 & y < 540) %>% 
	        	    dplyr::group_by(y) %>%
            	    dplyr::arrange(y, x) %>%
				    dplyr::mutate(text = trimws(text))
  
        	    d <- table_data %>%
            	    dplyr::filter(x < 150)%>% 
            	    dplyr::summarise(Descricao = paste(text, collapse = "~")) %>%
            	    dplyr::mutate(y = plyr::round_any(y, accuracy = 10, f = floor))  %>%
	        	    # Referencia
            	    dplyr::left_join(get_column_by_position_10(table_data, 150, 170, "MesRef"), by = "y") %>% #150, 170
	        	    # Qtd. Reg.
            	    dplyr::left_join(get_column_by_position_10(table_data, 170, 210, "QtdReg"), by = "y") %>% #170, 210
	        	    # Qtd. Fat.
            	    dplyr::left_join(get_column_by_position_10(table_data, 210, 250, "QtdFat"), by = "y") %>% #210, 250
	        	    # Unidade
            	    dplyr::left_join(get_column_by_position_10(table_data, 250, 270, "Unid"), by = "y") %>% #250, 270
	        	    # tarifa com tributos
            	    dplyr::left_join(get_column_by_position_10(table_data, 270, 310, "Tartri"), by = "y") %>% #270, 310
	        	    # vlr Tot
            	    dplyr::left_join(get_column_by_position_10(table_data, 310, 350, "VlrTot"), by = "y") %>% #310, 350
	        	    # base ICMS
            	    dplyr::left_join(get_column_by_position_10(table_data, 350, 390, "BaseCalc"), by = "y") %>% #350, 390
	        	    # aliqota ICMS
            	    dplyr::left_join(get_column_by_position_10(table_data, 390, 410, "AliqICMS"), by = "y") %>% #390, 410
	        	    # VlrICMS
            	    dplyr::left_join(get_column_by_position_10(table_data, 410, 440, "VlrICMS"), by = "y") %>% #410, 460 ALTERADO
	        	    # base pis & confins
            	    dplyr::left_join(get_column_by_position_10(table_data, 440, 490, "BasePisCof"), by = "y") %>% #460, 490 ALTERADO
	        	    # pis
            	    dplyr::left_join(get_column_by_position_10(table_data, 490, 520, "Pis"), by = "y") %>% #490, 520
            	    # confins
            	    dplyr::left_join(get_column_by_position_10(table_data, 520, 550, "Cofins"), by = "y") %>% #520, 550
				    # remove header line
	        	    dplyr::filter(y > 300) %>%
				    # remove coordinates
				    dplyr::select(-y)
			    # prepare pdf data in simplified way
			    pdf <- pdfs[[1]] %>% 
		    	    dplyr::group_by(y)  %>% 
		    	    dplyr::summarise(lines = paste(text, collapse = "~")) 
				# Get referencia
			    Referencia <- pdf %>%
				    dplyr::mutate(lines = stringr::str_extract(lines, pattern = "[A-Z]{3}/\\d{4}.+\\d{2}/\\d{2}/\\d{4}.+"),
                          	      lines = stringr::str_extract(lines, pattern = "[A-Z]{3}/\\d{4}")) %>%
				    dplyr::filter(!is.na(lines)) %>%
				    dplyr::pull(lines)
			    # get emission and presentation date
			    DtEmis <- get_regex_2(pdf, "Data~de~Emiss", "\\d{2}/\\d{2}/\\d{4}")
			    DtApre <- get_regex_2(pdf, "Data~de~Apresenta", "\\d{2}/\\d{2}/\\d{4}")
			    # extract simplified data from second page of pdf
        	    meta_data <- pdfs[[2]] %>% 
		    	    dplyr::group_by(y)  %>% 
		    	    dplyr::summarise(lines = paste(text, collapse = "~")) 
			    # get dates of measurement
			    leituras <- meta_data %>%
		    	    dplyr::filter(stringr::str_detect(stringr::str_to_lower(lines), pattern = "^leitura.*\\d{2}/\\d{2}/\\d{4}")) %>%
				    dplyr::transmute(lines = trimws(stringr::str_remove(stringr::str_to_lower(lines), "leitura~"))) %>%
				    tidyr::separate(lines, into = c("tipo","data"), sep = "~") %>%
				    tidyr::pivot_wider(names_from = tipo, values_from = data)
			    # get quantity of days
			    dias <- meta_data %>%
		    	    dplyr::filter(stringr::str_detect(stringr::str_to_lower(lines), pattern = "^qtd~de~dias.+\\d{1,2}")) %>%
				    dplyr::transmute(lines = trimws(stringr::str_remove(stringr::str_to_lower(lines), "^qtd~de~dias"))) %>%
				    dplyr::pull(lines) %>%
				    readr::parse_number()
			    # get unidade consumidora
			    UC <- meta_data %>%
		    	    dplyr::filter(stringr::str_detect(stringr::str_to_lower(lines), pattern = "^instala.+\\d{5,}")) %>%
				    dplyr::pull(lines) %>%
				    readr::parse_number()
				# 
				C_uni <- get_regex_3(meta_data, "nica~(\\d*\\.*\\d*\\.*\\d*)")
				C_pt <- get_regex_3(meta_data, "Ponta~(\\d*\\.*\\d*\\.*\\d*)(~|$)")
				C_fpt <- get_regex_3(meta_data, "Fora~Ponta~(\\d*\\.*\\d*\\.*\\d*)(~|$)")
        	    # build output data frame
        	    DadosTMP <- d %>%
            	    dplyr::mutate(UC = UC,
			              	      Referencia = Referencia,
                          	      LeitAnterior = leituras[[1, "anterior"]],
                          	      LeitAtual = leituras[[1, "atual"]],
						  	      DiasPeriodo = dias,
                          	      DtEmis = DtEmis,
						  	      DtApre = DtApre,
								  C_uni = C_uni,
								  C_pt = C_pt,
								  C_fpt = C_fpt) %>%
		    	    dplyr::mutate(across(c(3, 4, 6:13, 18), ~stringr::str_remove_all(.x, pattern = "\\.") %>% 
		                                                         stringr::str_replace(pattern = ",", replacement = "\\.") %>%
													             trimws()%>%
													             stringr::str_replace(pattern = '(\\d+\\.\\d+)(-)$', replacement = '\\2\\1') %>%
													             readr::parse_number()),
		                          across(c(16,17,19,20), ~lubridate::dmy(.x))) %>% 
                    # split column if there is a number at the end
                    tidyr::extract(Descricao, into = c("CCI", "DescProd"), regex = "(^\\d{4})?(.+)", remove = TRUE, convert = FALSE)  %>%
			        dplyr::mutate(DescProd = stringr::str_replace_all(stringr::str_remove(DescProd, pattern = "^~"), pattern = "~", replacement = " ")) 
		
                return(DadosTMP)
            }     
            # make it error proof and the error recognizable in result data
            explore_fatura_a <- purrr::possibly(explore_fatura_a, otherwise = data.frame(CCI = NA,
				                                                                         DescProd = "0",
	                                                                                     MesRef = NA,
																				         QtdReg = NA,
                                                                                         QtdFat = NA,
                                                                                         Unid = NA,
                                                                                         Tartri = NA,
                                                                                         VlrTot = NA,
                                                                                         BaseCalc = NA,
                                                                                         AliqICMS = NA,
                                                                                         VlrICMS = NA,
                                                                                         BasePisCof = NA,
																				         Pis = NA,
                                                                                         Cofins = NA,
																				         UC = NA,
																				         Referencia = NA,
																				         LeitAnterior = NA,
																				         LeitAtual = NA,
																				         DiasPeriodo = NA,
																				         DtEmis = NA,
																				         DtApre = NA,
																						 C_uni = NA,
																						 C_pt = NA,
																						 C_fpt = NA))
            # apply function on vector
            aret <- purrr::map(grupo_a, ~explore_fatura_a(.x)) 
	        # name list result
            names(aret) <- grupo_a
	        # bind to data frame and solve reference date
	        aret <- plyr::ldply(aret) %>%
				dplyr::mutate(Referencia = lubridate::my(paste(match(stringr::str_to_lower(stringr::str_sub(Referencia, start = 1, end = 3)), meses), "-", stringr::str_sub(Referencia, start = 5))))
	        # if no prior data exists create empty df	
	        if(is.null(read_data_A)){
	            read_data_A <- aret[0, ] 
	        }
	        # detect process errors
            erro_de_proces_a <- data.frame(path = aret[aret$DescProd == "0", 1]) %>%
                dplyr::mutate(motivo = "erro fatura A") %>%
		        dplyr::filter(!is.na(path))
	        # remove process error from fatura data output
	        aret <- aret %>%
	            dplyr::filter(DescProd != "0" & !is.na(VlrTot)) 
	        # secure against all error result
	        if(nrow(aret) == 0){
	            aret <- read_data_A[0, ] 
	        }
			# data verification	current data detect faturation type
			aavisos1 <- get_descprod_with_info_CPFL_A(aret, sear = "consumo reativo exc fora ponta", info = "consumo reativo exc fora ponta") %>%
			    dplyr::union(get_descprod_with_info_CPFL_A(aret, sear = "consumo reativo exc ponta", info = "consumo reativo exc ponta")) %>%
			    dplyr::union(get_descprod_with_info_CPFL_A(aret, sear = "demanda reat exced.+tusd", info = "demanda reat exced")) %>%
			    dplyr::union(get_descprod_with_info_CPFL_A(aret, sear = "demanda ultrap.+tusd", info = "demanda ultrap")) %>%
			    dplyr::union(get_descprod_with_info_CPFL_A(aret, sear = "demanda ultrap fponta.+tusd", info = "demanda ultrap fponta")) %>%
			    dplyr::union(get_descprod_with_info_CPFL_A(aret, sear = "demanda ultrap ponta.+tuds", info = "demanda ultrap ponta")) %>%
			    dplyr::union(get_descprod_with_info_CPFL_A(aret, sear = "multa por atraso pgto", info = "multa por atraso pgto"))
	        # bind old and new data
            aret <- aret %>%
	            dplyr::mutate(AliqICMS = AliqICMS / 100) %>%
		        dplyr::union_all(read_data_A) 
			# data verification	current data get unit with growth in last 3
			aavisos2 <- aret %>% 
				    dplyr::group_by(UC, Referencia) %>% 
					dplyr::summarize(QtdReg = sum(QtdReg, na.rm = TRUE)) %>% 
					dplyr::arrange(UC, Referencia) %>% 
					dplyr::slice_max(order_by = Referencia, n = 3) %>% 
					dplyr::mutate(dif = QtdReg - lead(QtdReg)) %>%
					dplyr::filter(dif > 0) %>%
					dplyr::count() %>%
					dplyr::filter(n == 2) %>%
					dplyr::transmute(Grupo = "A",
					                 UC = as.character(UC))
	        # get fatura summary errors
            erro_vlr_total_a <- aret %>% 
    	        dplyr::filter(!stringr::str_detect(CCI, pattern = "^09") & ! stringr::str_detect(stringr::str_to_lower(DescProd), pattern = "(^total)|(^subtotal)")) %>% 
		        dplyr::group_by(UC, .id) %>%
		        dplyr::summarize(SomTot = sum(VlrTot)) %>%
		        dplyr::ungroup() %>% 
		        dplyr::inner_join(aret %>% 
                          	          dplyr::filter(stringr::str_detect(DescProd, pattern = "^Subtotal")) %>%
	                      	          dplyr::select(UC, .id, VlrTot), by = c("UC", ".id")) %>%
		        dplyr::filter(abs(as.numeric(VlrTot) - as.numeric(SomTot))>.001) %>%
		        dplyr::transmute(path = .id,
	                	         motivo = "erro valor total") %>%
		        dplyr::distinct()
        }else{
	        # set new data equal to old data as there no new fatura of this type
            aret <- read_data_A
	        erro_vlr_total_a <- NULL
	        erro_de_proces_a <- NULL
        }
	},error = function(e){ 
    msgBox <- tcltk::tkmessageBox(title = "PROBLEMA ;(",
                                  message = "Algo causou falha critica no processamento das faturas do tipo A", icon = "info", type = "ok")
    UNDECLARED()
  }
)		
####################################################################
# only process type B if any is there to read
tryCatch(
	expr = {
        if(length(grupo_b) != 0){
            # main function to extract data
            explore_fatura_b <- function(arquivo){	
                # read data from pdf file
		        pdfs <- pdftools::pdf_data(arquivo)
		        # filter table region from first page of pdf doc 
		        table_data <- pdfs[[1]] %>% 
                    dplyr::filter(y > 320 & y < 480) %>% 
	                dplyr::group_by(y) %>%
                    dplyr::arrange(y, x) %>%
			        dplyr::mutate(text = trimws(text))

                d <- table_data %>%
                    dplyr::filter(x < 180)%>% 
                    dplyr::summarise(Descricao = paste(text, collapse = "~")) %>%
                    dplyr::mutate(y = plyr::round_any(y, accuracy = 5, f = floor))  %>%
	                # 3. Referencia
                    dplyr::left_join(get_column_by_position_5(table_data, 180, 190, "MesRef"), by = "y") %>%
	                # 4. Qtd. Reg.
                    dplyr::left_join(get_column_by_position_5(table_data, 190, 225, "QtdFat"), by = "y") %>%
	                # 5. Qtd. Fat.
                    dplyr::left_join(get_column_by_position_5(table_data, 225, 240, "Unid"), by = "y") %>%
	                # 6. Tarifa com tributos
                    dplyr::left_join(get_column_by_position_5(table_data, 240, 270, "Tartri"), by = "y") %>%
	                # 7. vlr Tot
                    dplyr::left_join(get_column_by_position_5(table_data, 270, 315, "VlrTot"), by = "y") %>%
			        # 8. Base Calculo ICMS
                    dplyr::left_join(get_column_by_position_5(table_data, 315, 360, "BaseCalc"), by = "y") %>%
	                # 9. Aliqota ICMS
                    dplyr::left_join(get_column_by_position_5(table_data, 360, 380, "AliqICMS"), by = "y") %>% 
	                # 10. VlrICMS
                    dplyr::left_join(get_column_by_position_5(table_data, 380, 410, "VlrICMS"), by = "y") %>%
	                # 11. base pis & confins
                    dplyr::left_join(get_column_by_position_5(table_data, 410, 450, "BasePisCof"), by = "y") %>%
	                # 12. pis
                    dplyr::left_join(get_column_by_position_5(table_data, 450, 480, "Pis"), by = "y") %>%
                    # 13. confins
                    dplyr::left_join(get_column_by_position_5(table_data, 480, 510, "Cofins"), by = "y") %>%
			        # remove header line
	                dplyr::filter(y > 330) %>%
			        # remove coordinates
			        dplyr::select(-y)
			    # prepare pdf data in simplified way
		        pdf <- pdfs[[1]] %>%
		            dplyr::group_by(y) %>% 
		            dplyr::summarise(lines = paste0(text, collapse = " "))
			    # Get referencia
		        Referencia <- pdf %>%
			        dplyr::mutate(lines = stringr::str_extract(lines, pattern = "[A-Z]{3}/\\d{4}.+\\d{2}/\\d{2}/\\d{4}.+"),
                                  lines = stringr::str_extract(lines, pattern = "[A-Z]{3}/\\d{4}")) %>%
                    dplyr::filter(!is.na(lines)) %>%
                    dplyr::pull(lines)
                # get specific info by regex with custom function
                UC <- get_regex_2(pdf, "^www\\.cpfl\\.com\\.br", "\\d{5,}")[1]
		        # get emission and presentation date
		        DtEmis <- get_regex_2(pdf, "Data de Emiss", "\\d{2}/\\d{2}/\\d{4}")
		        DtApre <- get_regex_2(pdf, "Data de Apresenta", "\\d{2}/\\d{2}/\\d{4}") 
                # get dates of measurement	
		        LeiTitle <- pdf %>%
		                        dplyr::filter(stringr::str_detect(stringr::str_to_lower(lines), pattern = "leitura.+leitura")) %>%
						        dplyr::pull(y)		
		        LeitData <- pdf %>%
					            dplyr::filter(y > LeiTitle & stringr::str_detect(lines, pattern = "kWh.+%")) %>%
						        dplyr::pull(lines) %>%
						        stringr::str_extract("\\d{2}/\\d{2}/\\d{4}\\s+\\d{2}/\\d{2}/\\d{4}") %>%
						        strsplit(split = "\\s") %>%
						        .[[1]]
		        LeitAtual <- LeitData[1]
		        LeitAnterior <- LeitData[2]
		        # build output data frame
                DadosTMP <- d %>%
                    dplyr::mutate(UC = UC,
			                      Referencia = Referencia,
                                  LeitAnterior = LeitAnterior,
                                  LeitAtual = LeitAtual,
                                  DtEmis = DtEmis,
						          DtApre = DtApre) %>%
		            dplyr::mutate(across(c(3, 5:12), ~stringr::str_remove_all(.x, pattern = "\\.") %>% 
		                                                  stringr::str_replace(pattern = ",", replacement = "\\.") %>%
											              trimws()%>%
											              stringr::str_replace(pattern = '(\\d+\\.\\d+)(-)$', replacement = '\\2\\1') %>%
											              readr::parse_number()),
		                          across(15:18, ~lubridate::dmy(.x))) %>% 
                    # split column if there is a number at the end
                    tidyr::extract(Descricao, into = c("CCI", "DescProd"), regex = "(^\\d{4})?(.+)", remove = TRUE, convert = FALSE) %>%
			        dplyr::mutate(DescProd = stringr::str_replace_all(stringr::str_remove(DescProd, pattern = "^~"), pattern = "~", replacement = " ")) 
		        # explicite return
		        return(DadosTMP)
	        }
            # make it error proof and the error recognizable in result data
            explore_fatura_b <- purrr::possibly(explore_fatura_b, otherwise = data.frame(CCI = NA, 
	                                                                                     DescProd = "0",
	                                                                                     MesRef = NA,
							        													 QtdFat = NA,
                                                                                         Unid = NA,
                                                                                         Tartri = NA,
                                                                                         VlrTot = NA,
															        					 BaseCalc = NA,
                                                                                         AliqICMS = NA,
                                                                                         VlrICMS = NA,
                                                                                         BasePisCof = NA,
					        															 Pis = NA,
                                                                                         Cofins = NA,
									        											 UC = NA,
											        									 Referencia = NA,
													        							 LeitAnterior = NA,
															        					 LeitAtual = NA,
																	        			 DtEmis = NA,
																			        	 DtApre = NA))
	        # apply custom function on vector
            bret <- purrr::map(grupo_b, ~explore_fatura_b(.x))
	        # name list result
            names(bret) <- grupo_b
	        # bind to data frame and solve reference date
	        bret <- plyr::ldply(bret) %>%
				dplyr::mutate(Referencia = lubridate::my(paste(match(stringr::str_to_lower(stringr::str_sub(Referencia, start = 1, end = 3)), meses), "-", stringr::str_sub(Referencia, start = 5))))
	        # if no prior data exists create empty df	
	        if(is.null(read_data_B)){
	            read_data_B <- bret[0, ] 
	        }
            # detect process errors
            erro_de_proces_b <- data.frame(path = bret[bret$DescProd == "0", 1]) %>%
                dplyr::mutate(motivo = "erro fatura B") %>%
	        	dplyr::filter(!is.na(path))	
	        # remove process error from fatura data output
	        bret <- bret %>%
	            dplyr::filter(DescProd != "0" & !is.na(VlrTot)) 
	        #
	        if(nrow(bret) == 0){
	            bret <- read_data_B[0, ] 
	        }
	        # remove process error from fatura data output
            bret <- bret %>%
	            dplyr::mutate(AliqICMS = AliqICMS / 100) %>%
		        dplyr::union_all(read_data_B) 
			#
			bavisos2 <- bret %>% 
				dplyr::group_by(UC, Referencia) %>% 
				dplyr::summarize(QtdFat = sum(QtdFat, na.rm = TRUE)) %>% 
				dplyr::arrange(UC, Referencia) %>% 
				dplyr::slice_max(order_by = Referencia, n = 3) %>% 
				dplyr::mutate(dif = QtdFat - lead(QtdFat)) %>%
				dplyr::filter(dif > 0) %>%
				dplyr::count() %>%
				dplyr::filter(n == 2) %>%
				dplyr::transmute(Grupo = "B",
				                 UC = as.character(UC))
			
	        # get fatura summary errors
	        erro_vlr_total_b <- bret %>% 
    	        dplyr::filter(!stringr::str_detect(CCI, pattern = "^09") & ! stringr::str_detect(stringr::str_to_lower(DescProd), pattern = "(^total)|(^subtotal)")) %>%  
		        dplyr::group_by(UC, .id) %>%
		        dplyr::summarize(SomTot = sum(VlrTot)) %>%
		        dplyr::ungroup() %>% 
		        dplyr::inner_join(bret %>% 
                          	          dplyr::filter(stringr::str_detect(stringr::str_to_lower(DescProd), pattern = "total distribuido")) %>%
	                      	          dplyr::select(UC, .id, VlrTot), by = c("UC", ".id")) %>%
		        dplyr::filter(abs(as.numeric(VlrTot) - as.numeric(SomTot))>.001) %>%
		        dplyr::transmute(path = .id,
	                	         motivo = "erro valor total") %>%
		        dplyr::distinct()
        }else{
	        # set new data equal to old data as there no new fatura of this type
            bret <- read_data_B
	        erro_vlr_total_b <- NULL
	        erro_de_proces_b <- NULL
        }
	},error = function(e){ 
    msgBox <- tcltk::tkmessageBox(title = "PROBLEMA ;(",
                                  message = "Algo causou falha critica no processamento das faturas do tipo B", icon = "info", type = "ok")
    UNDECLARED()
  }
)
# data combination and export
tryCatch(
	expr = {		
        # create wb object
        wb <- openxlsx::createWorkbook()
		# create worksheets
        openxlsx::addWorksheet(wb, "novidades")
		        # Create a percent style
        pct <- openxlsx::createStyle(numFmt="0%")
        # add percentage style to columns on specific sheets
        if(!is.null(aret)){
            openxlsx::addWorksheet(wb, "fat_A")
            openxlsx::writeData(wb, "fat_A", aret)
			openxlsx::addStyle(wb, "fat_A", style = pct, rows = 1:nrow(aret), cols = 11, gridExpand = TRUE)
		}
        if(!is.null(bret)) {
            openxlsx::addWorksheet(wb, "fat_B")
            openxlsx::writeData(wb, "fat_B", bret)	
            openxlsx::addStyle(wb, "fat_B", style = pct, rows = 1:nrow(bret), cols = 10, gridExpand = TRUE)
        }
		
        openxlsx::addWorksheet(wb, "avisos")
		openxlsx::addWorksheet(wb, "crescente")
        # write data to wb object by name
        openxlsx::writeData(wb, "novidades", rbind(segunda_via, nao_e_fatura, erro_de_proces_a, erro_de_proces_b, erro_vlr_total_a, erro_vlr_total_b) %>%
                             dplyr::select(motivo, path))

		if(exists("aavisos1")) openxlsx::writeData(wb, "avisos", aavisos1)
		if(exists("aavisos2") & exists("bavisos2")){
			openxlsx::writeData(wb, "crescente", dplyr::union(aavisos2, bavisos2))
		} else if (exists("aavisos2") & !exists("bavisos2")) {
			openxlsx::writeData(wb, "crescente", aavisos2)
		} else if (!exists("aavisos2") & exists("bavisos2")) {
			openxlsx::writeData(wb, "crescente", bavisos2)
		}
        # save object to excel file
        openxlsx::saveWorkbook(wb, Excel_Output, overwrite = TRUE)
        # final message: if you got here all went well
        tcltk::tkmessageBox(title = "SOLUCIONADO :)", message = paste("O arquivo output foi criado ou atualizado!", Excel_Output), icon = "info", type = "ok")
	},error = function(e){ 
    msgBox <- tcltk::tkmessageBox(title = "PROBLEMA ;(",
                                  message = "A construcao do resumo ou exportacao causaram o erro", icon = "info", type = "ok")
    UNDECLARED()
  }
)